/*
** file: mpchc.c
**
** UC: 21111 - Sistemas Operativos
** e-folio A 2016-17
**
** Cadeia de np processos com o mesmo FILE*
** Troca maiúsculas <--> minúsculas (Multiprocess CHange Case)
** Versão demo - uma de muitas soluções possíveis!
**  
** Utilização: mpchc np file1 file2
**
** Compilar: gcc -Wall -o mpchc mpchc.c
**
** Paulo Shirley @ UAb
** v2016/04/26
*/

/* NOTA: Este programa implementa 2 estratégias possíveis em 2 if's. Ativar o
if correspondente com if( 1 ) */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

void processa(int i, int np, FILE *fp1, FILE *fp2, long n, long nc, int b);


int main(int argc, char *argv[])
{
   pid_t pid;
   int status,i,np;
   long n,nc;
   FILE *fp1,*fp2;


   /* verificar numero argumentos */
   if( argc!=4 ) {
      printf("\n Utilizacao:  mpchc np ficheiro1 ficheiro2\n\n");
      exit(1);
   }
   /* verificar num. processos */
   np= atoi(argv[1]);
   if( np<1 ) {
      printf("\n Erro em np\n\n");
      exit(1);
   }
   /* verificar ficheiro1 de entrada */
   fp1=fopen(argv[2],"r");
   if( fp1==NULL ) {
      printf("\n Erro no ficheiro1\n\n");
      exit(1);
   }
   /* abrir ficheiro2 de saida */
   fp2=fopen(argv[3],"w");
   if( fp2==NULL ) {
      printf("\n Erro no ficheiro2\n\n");
      exit(1);
   }
   /* determinar comprimento do ficheiro entrada */
   fseek(fp1,0L,SEEK_END);
   n= ftell(fp1);
   nc= n/np;      /* num caracteres para cada processo, excepto último */

   /* print info inicial (opcional) */
   printf("\n\nTroca de letras minusculas <--> maiusculas\nficheiro %s para o" \
      " ficheiro %s\n%d processos, blocos de %ld e total de %ld carateres\n\n",\
      argv[2],argv[3],np,nc,n);

   /* Estratégia 1: Processo 1 é o 1º a processar um bloco do ficheiro.
   Pai processa e depois faz fork-wait, filho só é criado depois do trabalho
   do pai estar completo. Último processo só processa. */
   if ( 1 ) {
      for(i=1; i<=np; ++i) {
         processa(i,np,fp1,fp2,n,nc,i);  /* pai processa antes de criar filho */
         if( i<np ) {
            fflush(fp2);           /* garante escrita buffers antes do fork() */
            pid= fork();
            if( pid==-1 ) exit(1); /* erro fork: termina */
            if( pid>0 ) {
               wait(&status);      /* pai espera pelo filho e termina */
               break;
            }
            /* filho avança proximo i, processa e torna-se pai */
         }
      }
   }

   /* Estratégia 2: Processo np é o 1º a processar um bloco do ficheiro.
   Filho processa e depois termina para wait do pai, pai só processa depois do
   trabalho do filho estar completo. Inverte bloco a processar com j=np-i+1 para
   manter o processamento do início para o fim do ficheiro (opcional) */
   if ( 0 ) {
      for(i=1; i<=np; ++i) {
         if( i<np ) {
            pid= fork();
            if( pid==-1 ) exit(1);  /* erro fork: termina */
            if( pid>0 ) {
               wait(&status);   /* pai espera pelo filho e só depois processa */
               processa(i,np,fp1,fp2,n,nc,np-i+1);
               break;
            }
            /* filho avança proximo i, torna-se pai e faz fork */
         } else
            processa(i,np,fp1,fp2,n,nc,np-i+1);   /* ultimo filho só processa */
      }
   }

   fclose(fp1);
   fclose(fp2);
   return 0;
}


void processa(int i, int np, FILE *fp1, FILE *fp2, long n, long nc, int b)
/* i: processo, b: bloco a processar */
{
   long p0,pf,nb,k;
   int c;

   p0= (b-1)*nc;               /* início bloco a processar */
   pf= b==np ? n-1 : b*nc-1;   /* fim bloco */
   nb= pf-p0+1;                /* dimensão do bloco */
   if( nb>0 ) {
      fseek(fp1,p0,SEEK_SET);  /* garante posicionamento dos ficheiros */
      fseek(fp2,p0,SEEK_SET);
      for(k=0 ; k<nb; ++k) {   /* processa bloco i */
         c= fgetc(fp1);
         if ( islower(c) )
            c= toupper(c);
         else if( isupper(c) )
            c= tolower(c);
         fputc(c,fp2);
      }
   }
   printf("Processo %d: PID=%d, PPID=%d, convertidos %ld caracteres de %ld a" \
      " %ld\n",i,(int)getpid(),(int)getppid(),nb,p0,pf);
}


/* EOF */

